package;

import format.png.Data;
import format.png.Tools;
import format.png.Writer;
import haxe.io.Bytes;
import haxe.io.BytesInput;
import haxe.io.BytesOutput;
import haxe.io.Eof;
import haxe.io.Input;
import haxe.io.Path;
import sys.FileSystem;
import sys.io.File;
import sys.io.FileInput;
import sys.io.FileOutput;
import sys.io.FileSeek;

/**
 * NUKA-CEL! by Athena/ユウリ マヤ
 */
class Main
{
	private var path:String;
	private var countART2:Int;
	private var countCO2R:Int;
	private var countCLC2:Int;
	private var countSDL:Int;
	
	static public function main( )
	{
		new Main( );
	}

	public function new()
	{
		this.countART2 = 0;
		this.countCO2R = 0;
		this.countCLC2 = 0;
		this.countSDL = 0;

		var args = Sys.args( );
		if ( args.length == 0 ) {
			trace( "usage: nukacel [file]" );
			return;
		}

		var file = args[0];
		this.path = Path.withoutExtension( Path.withoutDirectory( file ) );

		FileSystem.createDirectory( path );
		FileSystem.createDirectory( path+"/ART2" );
		FileSystem.createDirectory( path+"/CO2R" );
		FileSystem.createDirectory( path+"/CLC2" );
		// FileSystem.createDirectory( path+"/SDL" );

		// 開く
		var f:FileInput = File.read( file );
		f.bigEndian = false;

		// 無理矢理解析
		while ( true ) {
			var pos = f.tell( );
			if ( pos & 3 != 0 ) {
				trace( "4 byte align ni teisei" );
				f.seek( -(pos & 3), FileSeek.SeekCur );
			}
			try {
				var header = f.readString( 4 );
				switch( header ) {
					case "ART2":
						this.extractART2( f );
					case "CO2R":
						this.extractCO2R( f );
					case "CLC2":
						this.extractCLC2( f );
					// case "SDL ":
					//	this.extractSDL( f );
				}
			}catch ( eof:Eof ) {
				// おわりなので抜ける
				break;
			}catch ( d:Dynamic ) {
				// 表示だけして握り潰す
				trace( "address:" + f.tell( ), d );
			}
		}

		f.close( );

		trace( "done!" );
	}

	/**
	 * ART2を抽出する（PNGとして）
	 * @param	fi
	 */
	private function extractART2( f:FileInput )
	{
		var bitDepth = f.readInt32( );
		if ( bitDepth < 0 ) {
			throw 'bitdepth minus toka okashii!!!! chuushi!!!!!!!!';
		}
		var width = f.readInt32( );
		var height = f.readInt32( );
		if ( ( width < 0 ) || ( height < 0 )||( 1024 <= width ) || ( 1024 <= height )  ) {
			throw 'gazou size ga hentekorin!!!!!! chuushi!!!!!!!!';
		}
		var name = f.readString( 0x10 ).split( String.fromCharCode( 0x00 ) )[0];
		var len = switch( bitDepth ) {
			case 4: ( width >> 1 ) * height;
			case 8: width * height;
			default:
				throw '$name: bit depth ga hen janai??? size:$bitDepth';
		}

		var imgPos = f.tell( );

		// 先にパレットをとっとく
		f.seek( len, FileSeek.SeekCur );
		var palette = this.extractPallete( f );

		// 画像を抜く
		f.seek( imgPos, FileSeek.SeekBegin );
		var bo = new BytesOutput( );
		bo.bigEndian = true;
		switch( bitDepth ) {
			case 4:
				for ( i in 0 ... len ) {
					var b = f.readByte( );
					bo.writeInt32( palette[ b & 0x0f ] );
					bo.writeInt32( palette[ ( b >> 4 ) & 0x0f ] );
				}
			case 8:
				for ( i in 0 ... len ) {
					bo.writeInt32( palette[f.readByte()] );
				}
		}
		f.seek( 0x100 * 4, FileSeek.SeekCur );

		// PNGとして保存
		var fo = File.write( this.path + "/ART2/" + this.countART2 + ".png" );
		new Writer(fo).write( Tools.build32ARGB(width, height, bo.getBytes( ) ) );
		fo.close( );
		trace( "ART2 get daze: " + name + " : " + this.countART2 );
		this.countART2 ++;
	}

	/**
	 * CO2Rを抽出（PNGとして）
	 * @param	fi
	 */
	private function extractCO2R( f:FileInput )
	{
		f.seek( 0x10, FileSeek.SeekCur );
		var count = f.readInt32( );
		if ( 256 <= count ) {
			throw "meccha gazou ga ooikara hen dato omou!!!!!! dakara chuushi!!!!!!!!!!!!!!!! " + count;
		}
		if ( count < 0 ) {
			throw "gazou maisuu ga minus dakara chuushi!!!!!!!!!!!!!!!! " + count;
		}

		// 名前
		f.seek( 8, FileSeek.SeekCur );
		var name = f.readString( 0x10 ).split( String.fromCharCode( 0x00 ) )[0];

		var dir = this.path + "/CO2R/" + this.countCO2R;
		FileSystem.createDirectory( dir );
		trace( "CO2R get daze: " + this.countCO2R );
		this.countCO2R ++;

		// 画像まで全部無視
		var p = f.tell( );
		while ( true ) {
			var magic = f.readString( 4 );
			if ( magic == "LTMR" ) break;
		}
		f.seek( 8, FileSeek.SeekCur );

		// 画像を全て取る
		for ( imageNumber in 0 ... count ) {
			var len = f.readInt32( );
			f.seek( 4, FileSeek.SeekCur );
			var bitDepth = f.readInt32( );
			var width = f.readInt32( );
			var height = f.readInt32( );
			f.seek( 12, FileSeek.SeekCur );
			var palette = this.extractPallete( f );

			// 画像を抜く
			var bo = new BytesOutput( );
			bo.bigEndian = true;
			switch( bitDepth ) {
				case 1:
					for ( i in 0 ... len ) {
						var b = f.readByte( );
						bo.writeInt32( palette[ b & 0x0f ] );
						bo.writeInt32( palette[ ( b >> 4 ) & 0x0f ] );
					}
				case 2:
					for ( i in 0 ... len ) {
						bo.writeInt32( palette[f.readByte()] );
					}
			}

			// PNGとして保存
			var fo = File.write( dir + "/" + imageNumber + ".png" );
			new Writer(fo).write( Tools.build32ARGB(width, height, bo.getBytes( ) ) );
			fo.close( );

			trace( '\t1 mai get daze: $imageNumber' );
		}
	}

	/**
	 * パレット取得
	 * @param	f
	 * @return
	 */
	private function extractPallete( f:FileInput ):Array<Int>
	{
		var palette:Array<Int> = [for ( i in 0 ... 0x100 ) 0];

		for ( i in 0 ... 0x100 ) {
			var r = f.readByte( );
			var g = f.readByte( );
			var b = f.readByte( );
			var a = Std.int( Math.min( f.readByte( ) << 1, 0xff ) );
			var argb = (a << 24) | ( r << 16 ) | ( g << 8 ) | b;
			palette[ ( i & 0xe7 ) | ( ( i & 0x10 ) >> 1 ) | ( ( i & 0x08) << 1 ) ] = argb;
		}

		return palette;
	}

	/**
	 * CLC2を抽出する
	 * @param	f
	 */
	private function extractCLC2( f:FileInput )
	{
		if ( f.readInt32( ) != 0x00000001 ) {
			throw "chigauppoi node chuushi!!!!";
		}
		if ( f.readInt32( ) != 0x00000000 ) {
			throw "chigauppoi node chuushi!!!!";
		}

		f.seek( 4, FileSeek.SeekCur );
		var size = f.readInt32( );
		if ( 0x1000000 < size ) {
			throw "ookisugiru node chuushi!!!!";
		}

		f.seek( -20, FileSeek.SeekCur );
		var b = f.read( 20 + size );

		// バイナリを保存
		var nowCLC2 = this.countCLC2;
		var fo = File.write( this.path + "/CLC2/" + nowCLC2 + ".clc2" );
		fo.write( b );
		fo.close( );
		trace( "CLC2 get daze: " + nowCLC2 );
		this.countCLC2 ++;

		// テキストを抽出する
		var allLines = new Array<String>( );
		var bi = new BytesInput( b );
		while ( bi.position < bi.length ) {
			// オペコード
			if ( bi.readInt32( ) != 0x000000cd ) continue;
			// 常に4
			if ( bi.readInt32( ) != 4 ) continue;
			// 64バイトとばす
			bi.position += 64;
			// テキストのサイズ
			var size = bi.readInt32( );
			if ( size < 0 ) {
				throw "text length ga minus toka nanka okashii!!!!!! chuushi!!!!!";
			}
			allLines.push( bi.readString( size ).split( String.fromCharCode( 0x00 ) )[0] );
		}

		if( 0 < allLines.length ) {
			var fo = File.write( this.path + "/CLC2/" + nowCLC2 + ".txt" );
			fo.writeString( allLines.join( "\r\n" ) );
			fo.close( );
		}
	}
}
